from copy import deepcopy
from itertools import tee
from math import pow, log10
import os
import sys

import numpy as np
import numpy.lib.recfunctions as nprf


def qstring_to_phred(quality):
    """Compute standard phred scores from a quality string."""
    qscores = [ord(q) - 33 for q in quality]
    return qscores


def mean_qscore(scores):
    """Returns the phred score corresponding to the mean of the probabilities
    associated with the phred scores provided. Taken from chimaera.common.utilities.


    :param scores: Iterable of phred scores.

    :returns: Phred score corresponding to the average error rate, as
        estimated from the input phred scores.
    """
    if len(scores) == 0:
        return 0.0
    sum_prob = 0.0
    for val in scores:
        sum_prob += pow(10, -0.1 * val)
    mean_prob = sum_prob / len(scores)
    return -10.0 * log10(mean_prob)


def kmer_overlap_gen(kmers, moves=None):
    """From a list of kmers return the character shifts between them.
    (Movement from i to i+1 entry, e.g. [AATC,ATCG] returns [0,1]).
    Allowed moves may be specified in moves argument in order of preference.
    Taken from dragonet.bio.seq_tools

    :param kmers: sequence of kmer strings.
    :param moves: allowed movements, if None all movements to length of kmer
        are allowed.
    """

    first = True
    yield 0
    for last_kmer, this_kmer in window(kmers, 2):
        if first:
            if moves is None:
                l = len(this_kmer)
                moves = range(l + 1)
            first = False

        l = len(this_kmer)
        for j in moves:
            if j < 0:
                if last_kmer[:j] == this_kmer[-j:]:
                    yield j
                    break
            elif j > 0 and j < l:
                if last_kmer[j:l] == this_kmer[0:-j]:
                    yield j
                    break
            elif j == 0:
                if last_kmer == this_kmer:
                    yield 0
                    break
            else:
                yield l
                break


def build_mapping_table(events, ref_seq, post, scale, path, model):
    """Build a mapping table based on output of a dragonet.mapper style object.
    Taken from chimaera.common.utilities.

    :param events: Numpy record array of events. Must contain the mean,
        stdv, start and length fields.
    :param ref_seq: String representation of the reference sequence.
    :param post: Numpy 2D array containing the posteriors (event, state).
    :param scale: Scaling object.
    :param path: Numpy 1D array containing position in reference. May contain
        negative values, which will be interpreted as "bad emissions".
    :param model: Model object to use.

    :returns: numpy record array containing summary fields. One record per event.

    ====================   =====================================================
    Output Field           Description
    ====================   =====================================================
    *mean*                 mean value of event samples (level)
    *scaled_mean*          *mean* scaled to the bare level emission (mean/mode)
    *stdv*                 standard deviation of event samples (noise)
    *scaled_stdv*          *stdv* scaled to the bare stdv emission (mode)
    *start*                start time of event /s
    *length*               length of event /s
    *model_level*          modelled event level, i.e. the level emission
                           associated with the kmer *kmer*, scaled to the data
    *model_scaled_level*   bare level emission
    *model_sd*             modelled event noise, i.e. the sd emission associated
                           with the kmer *kmer*, scaled  to the data
    *model_scaled_sd*      bare noise emission
    *seq_pos*              aligned sequence position, position on Viterbi path
    *p_seq_pos*            posterior probability of states on Viterbi path
    *kmer*                 kmer identity of *seq_pos*
    *mp_pos*               aligned sequence position, position with highest
                           posterioir
    *p_mp_pos*             posterior probability of most probable states
    *mp_kmer*              kmer identity of *mp_kmer*
    *good_emission*        whether or not the HMM has tagged event as fitting
                           the model
    ====================   =====================================================

    """
    kmer_len = len(model['kmer'][0])

    kmer_index = seq_to_kmers(ref_seq, kmer_len)
    label_index = dict((j,i) for i,j in enumerate(model['kmer']))
    kmer_dtype = '|S{}'.format(kmer_len)

    column_names = ['mean', 'scaled_mean', 'stdv', 'scaled_stdv', 'start', 'length',
                    'model_level', 'model_scaled_level', 'model_sd', 'model_scaled_sd',
                    'p_seq_pos', 'p_mp_pos', 'seq_pos', 'mp_pos', 'move', 'good_emission',
                    'kmer', 'mp_kmer']
    column_types = [float] * 12 + [int] * 3 + [bool] + [kmer_dtype] * 2
    table = np.zeros(events.size, dtype=list(zip(column_names, column_types)))

    zero_start = events['start'] - events['start'][0]

    # Sequence position
    seq_pos = np.where(path >= 0, path, np.abs(path) - 1)
    seq_kmer = [kmer_index[x] for x in seq_pos]
    seq_kmer_i = [label_index[i] for i in seq_kmer]

    table['seq_pos'] = seq_pos
    table['kmer'] = seq_kmer
    table['p_seq_pos'] = post[range(post.shape[0]), seq_pos]
    table['move'] = np.ediff1d(seq_pos, to_begin=[0])
    # Highest posterior positions
    mp_pos = np.argmax(post, axis=1)
    table['mp_pos'] = mp_pos
    table['mp_kmer'] = [kmer_index[x] for x in mp_pos]
    table['p_mp_pos'] = post[range(post.shape[0]), table['mp_pos']]
    # The data
    for x in ('mean', 'start','length', 'stdv'):
        table[x] = events[x]
    # scaling data to model
    table['scaled_mean'] = (table['mean'] - scale.shift - scale.drift * zero_start) / scale.scale
    table['scaled_stdv'] = table['stdv'] / scale.scale_sd
    # The model
    table['model_scaled_level'] = model['level_mean'][seq_kmer_i]
    table['model_scaled_sd']  = model['sd_mean'][seq_kmer_i]
    # The model scaled to the data
    table['model_level'] = scale.shift + scale.drift * zero_start + scale.scale * table['model_scaled_level']
    table['model_sd'] = scale.scale_sd * table['model_scaled_sd']
    # Tag ignore and outlier states
    table['good_emission'] = [x >= 0 for x in path]
    return table


def build_mapping_summary_table(mapping_summary):
    """Build a mapping summary table

    :param mapping_summary: List of curr_map dictionaries

    :returns: Numpy record array containing summary contents. One record per array element of mapping_summary

    """
    # Set memory allocation for variable length strings
    # This works, but there must be a better way
    max_len_name = 1
    max_len_direction = 1
    max_len_seq = 1
    for summary_line in mapping_summary:
        len_name = len(summary_line['name'])
        if len_name > max_len_name:
            max_len_name = len_name

        len_direction = len(summary_line['direction'])
        if len_direction > max_len_direction:
            max_len_direction = len_direction

        len_seq = len(summary_line['seq'])
        if len_seq > max_len_seq:
            max_len_seq = len_seq

    column_names = ['name', 'direction', 'is_best', 'score', 'scale', 'shift', 'drift', 'scale_sd', 'var_sd', 'var', 'seq']
    column_types = ['|S{}'.format(max_len_name)] + ['|S{}'.format(max_len_direction)] + [bool] + [float] * 7 + ['|S{}'.format(max_len_seq)]

    table = np.zeros(len(mapping_summary), dtype=list(zip(column_names, column_types)))
    for table_line, summary_line, in zip(table,mapping_summary):
        table_line['name'] = summary_line['name']
        table_line['direction'] = summary_line['direction']
        table_line['score'] = summary_line['score']
        table_line['scale'] = summary_line['scale'].scale
        table_line['shift'] = summary_line['scale'].shift
        table_line['drift'] = summary_line['scale'].drift
        table_line['scale_sd'] = summary_line['scale'].scale_sd
        table_line['var_sd'] = summary_line['scale'].var_sd
        table_line['var'] = summary_line['scale'].var
        table_line['seq'] = summary_line['seq']

    table['is_best'] = False
    is_best = np.argmin([line['score'] for line in mapping_summary])
    table[is_best]['is_best'] = True

    return table


def create_basecall_1d_output(raw_events, scale, path, model, post=None):
    """Create the annotated event table and basecalling summaries similiar to chimaera.

    :param raw_events: :class:`np.ndarray` with fields mean, stdv, start and,
        length fields.
    :param scale: :class:`dragonet.basecall.scaling.Scaler` object (or object
        with attributes `shift`, `scale`, `drift`, `var`, `scale_sd`, `var_sd`,
        and `var_sd`.
    :param path: list containing state indices with respect to `model`.
    :param model: `:class:dragonet.util.model.Model` object.
    :param post: Two-dimensional :class:`np.ndarray` containing posteriors (event, state).
    :param quality_data: :class:np.ndarray Array containing quality_data, used to annotate events.

    :returns: A tuple of:

        * the annotated input event table
        * a dict of result
    """

    events = raw_events.copy()
    model_state = np.array([model[x]['kmer'] for x in path])
    raw_model_level = np.array([model[x]['level_mean'] for x in path])
    move = np.array(list(kmer_overlap_gen(model_state)))
    counts = np.bincount(move)
    stays = counts[0]
    skips = counts[2] if len(counts) > 2 else 0

    # Extend the event table
    read_start = events[0]['start']
    model_level = scale.shift + scale.scale * raw_model_level +\
                  scale.drift * (events['start'] - read_start)
    new_columns = ['model_state', 'model_level', 'move']
    column_data = [model_state, model_level, move]

    if post is not None:
        weights = np.sum(post, axis=1)
        new_columns.append('weights')
        column_data.append(weights)

    drop_first = set(new_columns) & set(events.dtype.names)
    events = nprf.drop_fields(events, drop_first)
    table = nprf.append_fields(events, new_columns, data=column_data, asrecarray=True)

    # Compile the results
    results = {
        'num_events': events.size,
        'called_events': events.size,
        'shift': scale.shift,
        'scale': scale.scale,
        'drift': scale.drift,
        'var': scale.var,
        'scale_sd': scale.scale_sd,
        'var_sd': scale.var_sd,
        'num_stays': stays,
        'num_skips': skips
    }

    return table, results


def create_mapping_output(raw_events, scale, path, model, seq, post=None, n_states=None, is_reverse=False, substates=False):
    """Create the annotated event table and summaries similiar to chimaera

    :param raw_events: :class:`np.ndarray` with fields `mean`, `stdv`, `start`,
        and `length` fields.
    :param scale: :class:`dragonet.basecall.scaling.Scaler` object (or object
        with attributes `shift`, `scale`, `drift`, `var`, `scale_sd`, `var_sd`,
        and `var_sd`.
    :param path: list containing state indices with respect to `model`.
    :param model: `:class:dragonet.util.model.Model` object.
    :param seq: String representation of the reference sequence.
    :param post: Two-dimensional :class:`np.ndarray` containing posteriors (event, state).
    :param is_reverse: Mapping refers to '-' strand (bool).
    :param substate: Mapping contains substates?

    :returns: A tuple of:
        * the annotated input event table,
        * a dict of result.
    """

    events = raw_events.copy()
    direction = '+' if not is_reverse else '-'
    has_post = True

    # If we don't have a posterior, pass a mock object
    if post is None:
        if n_states is None:
            raise ValueError('n_states is required if post is None.')
        has_post = False
        post = MockZeroArray((len(events), n_states))
    table = build_mapping_table(events, seq, post, scale, path, model)

    # Delete mocked out columns
    if not has_post:
        to_delete = ['p_seq_pos', 'mp_pos', 'mp_kmer', 'p_mp_pos']
        table = nprf.drop_fields(table, to_delete)

    if direction == '-':
        events['seq_pos'] = len(seq) - table['seq_pos']
        ref_start = table['seq_pos'][-1]
        ref_stop = table['seq_pos'][0]
    else:
        ref_start = table['seq_pos'][0]
        ref_stop = table['seq_pos'][-1]

    # Compute movement stats.
    _, stays, skips = compute_movement_stats(path)

    results = {
        'direction': direction,
        'reference': seq,
        'ref_start': ref_start,
        'ref_stop': ref_stop,
        'shift': scale.shift,
        'scale': scale.scale,
        'drift': scale.drift,
        'var': scale.var,
        'scale_sd': scale.scale_sd,
        'var_sd': scale.var_sd,
        'num_stays': stays,
        'num_skips': skips
    }

    return table, results


class MockZeroArray(np.ndarray):
    def __init__(self, shape):
        """Mock enough of ndarray interface to be passable as a posterior matrix
        to chimaera build_mapping_table

        :param shape: tuple, shape of array
      
        """
        self.shape = shape

    def argmax(self, axis=0):
        """Fake argmax values of an array."""
        return np.zeros(self.shape[1-axis], dtype=int)


def validate_event_table(table):
    """Check if an object contains all columns of a basic event array."""

    if not isinstance(table, np.ndarray):
        raise TypeError('Table is not a ndarray.')

    req_fields = ['mean', 'stdv', 'start', 'length']
    if not set(req_fields).issubset(table.dtype.names):
        raise KeyError(
            'Array does not contain fields for event array: {}, got {}.'.format(
                req_fields, table.dtype.names
            )
        )


def validate_model_table(table):
    """Check if an object contains all columns of a dragonet Model."""
    if not isinstance(table, np.ndarray):
        raise TypeError('Table is not a ndarray.')

    req_fields = ['kmer', 'level_mean', 'level_stdv', 'sd_mean', 'sd_stdv']
    if not set(req_fields).issubset(table.dtype.names):
        raise KeyError(
            'Object does not contain fields required for Model: {}, got {}.'.format(
                req_fields, table.dtype.names
            )
        )


def validate_scale_object(obj):
    """Check if an object contains all attributes of dragonet Scaler."""

    req_attributes = ['shift', 'scale', 'drift', 'var', 'scale_sd', 'var_sd']
    msg = 'Object does not contain attributes required for Scaler: {}'.format(req_attributes)
    assert all([hasattr(obj, attr) for attr in req_attributes]), msg


def compute_movement_stats(path):
    """Compute movement stats from a mapping state path

    :param path: :class:`np.ndarry` containing position in reference.
        Negative values are interpreted as "bad emissions".
    """

    vitstate_indices = np.where(path >= 0, path, np.abs(path) - 1)
    move = np.ediff1d(vitstate_indices, to_begin=0)
    counts = np.bincount(move)
    stays = counts[0]
    skips = counts[2] if len(counts) > 2 else 0

    return move, stays, skips


def seq_to_kmers(seq, length):
    """Turn a string into a list of (overlapping) kmers.

    e.g. perform the transformation:

    'ATATGCG' => ['ATA','TAT', 'ATG', 'TGC', 'GCG']

    :param seq: character string
    :param length: length of kmers in output

    :returns: A list of overlapping kmers
    """
    return [seq[x:x + length] for x in range(0, len(seq) - length + 1)]


def window(iterable, size):
    """Create an iterator returning a sliding window from another iterator

    :param iterable: Iterator
    :param size: Size of window

    :returns: an iterator returning a tuple containing the data in the window

    """
    assert size > 0, "Window size for iterator should be strictly positive, got {0}".format(size)
    iters = tee(iterable, size)
    for i in range(1, size):
        for each in iters[i:]:
            next(each, None)
    return list(zip(*iters))


def readtsv(fname, fields=None, **kwargs):
    """Read a tsv file into a numpy array with required field checking

    :param fname: filename to read. If the filename extension is
        gz or bz2, the file is first decompressed.
    :param fields: list of required fields.
    """

    if not file_has_fields(fname, fields):
        raise KeyError('File {} does not contain requested required fields {}'.format(fname, fields))

    for k in ['names', 'delimiter', 'dtype']:
        kwargs.pop(k, None)
    table = np.genfromtxt(fname, names=True, delimiter='\t', dtype=None, encoding='utf8', **kwargs)
    #  Numpy tricks to force single element to be array of one row
    return table.reshape(-1)


def docstring_parameter(*sub):
    """Allow docstrings to contain parameters."""
    def dec(obj):
        obj.__doc__ = obj.__doc__.format(*sub)
        return obj
    return dec


def med_mad(data, factor=None, axis=None, keepdims=False):
    """Compute the Median Absolute Deviation, i.e., the median
    of the absolute deviations from the median, and the median

    :param data: A :class:`ndarray` object
    :param factor: Factor to scale MAD by. Default (None) is to be consistent
        with the standard deviation of a normal distribution
        (i.e. mad( N(0,\sigma^2) ) = \sigma).
    :param axis: For multidimensional arrays, which axis to calculate over
    :param keepdims: If True, axis is kept as dimension of length 1

    :returns: a tuple containing the median and MAD of the data

    """
    if factor is None:
        factor = 1.4826
    dmed = np.median(data, axis=axis, keepdims=True)
    dmad = factor * np.median(abs(data - dmed), axis=axis, keepdims=True)
    if axis is None:
        dmed = dmed.flatten()[0]
        dmad = dmad.flatten()[0]
    elif not keepdims:
        dmed = dmed.squeeze(axis)
        dmad = dmad.squeeze(axis)
    return dmed, dmad


def mad(data, factor=None, axis=None, keepdims=False):
    """Compute the Median Absolute Deviation, i.e., the median
    of the absolute deviations from the median, and (by default)
    adjust by a factor for asymptotically normal consistency.

    :param data: A :class:`ndarray` object
    :param factor: Factor to scale MAD by. Default (None) is to be consistent
        with the standard deviation of a normal distribution
        (i.e. mad( N(0,\sigma^2) ) = \sigma).
    :param axis: For multidimensional arrays, which axis to calculate the median over.
    :param keepdims: If True, axis is kept as dimension of length 1

    :returns: the (scaled) MAD

    """
    _ , dmad = med_mad(data, factor=factor, axis=axis, keepdims=keepdims)
    return dmad


def file_has_fields(fname, fields=None):
    """Check that a tsv file has given fields

    :param fname: filename to read. If the filename extension is
        gz or bz2, the file is first decompressed.
    :param fields: list of required fields.

    :returns: boolean
    """

    # Allow a quick return
    req_fields = deepcopy(fields)
    if isinstance(req_fields, str):
        req_fields = [fields]
    if req_fields is None or len(req_fields) == 0:
        return True
    req_fields = set(req_fields)

    inspector = open
    ext = os.path.splitext(fname)[1]
    if ext == '.gz':
        inspector = gzopen
    elif ext == '.bz2':
        inspector = bzopen

    has_fields = None
    with inspector(fname, 'r') as fh:
        present_fields = set(fh.readline().rstrip('\n').split('\t'))
        has_fields = req_fields.issubset(present_fields)
    return has_fields


def get_changes(data, ignore_cols=None, use_cols=None):
    """Return only rows of a structured array which are not equal to the previous row.

    :param data: Numpy record array.
    :param ignore_cols: iterable of column names to ignore in checking for equality between rows.
    :param use_cols: iterable of column names to include in checking for equality between rows (only used if ignore_cols is None).

    :returns: Numpy record array.
    """
    cols = list(data.dtype.names)
    if ignore_cols is not None:
        for col in ignore_cols:
            cols.remove(col)
    elif use_cols is not None:
        cols = list(use_cols)
    changed_inds = np.where(data[cols][1:] != data[cols][:-1])[0] + 1
    changed_inds = [0] + [i for i in changed_inds]
    return data[(changed_inds,)]


def _clean(value):
    """Convert numpy numeric types to their python equivalents."""
    if isinstance(value, np.ndarray):
        if value.dtype.kind == 'S':
            return np.char.decode(value).tolist()
        else:
            return value.tolist()
    elif type(value).__module__ == np.__name__:
        conversion = value.item()
        if sys.version_info.major == 3 and isinstance(conversion, bytes):
            conversion = conversion.decode()
        return conversion
    elif sys.version_info.major == 3 and isinstance(value, bytes):
        return value.decode()
    else:
        return value


def _clean_attrs(attrs):
    return {_clean(k): _clean(v) for k, v in attrs.items()}


def _sanitize_data_for_writing(data):
    if isinstance(data, str):
        return data.encode()
    elif isinstance(data, np.ndarray) and data.dtype.kind == np.dtype(np.unicode):
        return data.astype('S')
    elif isinstance(data, np.ndarray) and len(data.dtype) > 1:
        dtypes = dtype_descr(data)
        for index, entry in enumerate(dtypes):
            type_check = entry[1]
            if isinstance(type_check, tuple):
                # an enum?
                return data
            if type_check.startswith('<U'):
                # numpy.astype can't handle empty string datafields for some
                # reason, so we'll explicitly state that.
                if len(entry[1]) <= 2 or (len(entry[1]) == 3 and
                                          entry[1][2] == '0'):
                    raise TypeError('Empty datafield {} cannot be converted'
                                    ' by np.astype.'.format(entry[0]))
                dtypes[index] = (entry[0], '|S{}'.format(entry[1][2:]))
        return data.astype(dtypes)
    return data


def _sanitize_data_for_reading(data):
    if sys.version_info.major == 3:
        if isinstance(data, bytes):
            return data.decode()
        elif isinstance(data, np.ndarray) and data.dtype.kind == 'S':
            return np.char.decode(data)
        elif isinstance(data, np.ndarray) and len(data.dtype) > 1:
            dtypes = list(dtype_descr(data))
            for index, entry in enumerate(dtypes):
                type_check = entry[1]
                if isinstance(type_check, tuple):
                    # an enum?
                    return data
                if entry[1].startswith('|S'):
                    # numpy.astype can't handle empty datafields for some
                    # reason, so we'll explicitly state that.
                    if len(entry[1]) <= 2 or (len(entry[1]) == 3 and
                                              entry[1][2] == '0'):
                        raise TypeError('Empty datafield {} cannot be converted'
                                        ' by np.astype.'.format(entry[0]))
                    dtypes[index] = (entry[0], '<U{}'.format(entry[1][2:]))

                if entry[1].startswith('|u'):
                    # there seems to be a bug in astype
                    dtypes[index] = (entry[0], entry[1].replace('|u', 'u'))

            return data.astype(dtypes)
    return data


def dtype_descr(arr):
    """Get arr.dtype.descr
    Views of structured arrays in which columns have been re-ordered nolonger support arr.dtype.descr
    see https://github.com/numpy/numpy/commit/dd8a2a8e29b0dc85dca4d2964c92df3604acc212
    """
    try:
        return arr.dtype.descr
    except ValueError:
        return tuple([(n, arr.dtype[n].descr[0][1]) for n in arr.dtype.names])


def group_vector(arr):
    """Group a vector by unique values.

    :param arr: input vector to be grouped.

    :returns: a dictionary mapping unique values to arrays of indices of the
        input vector.

    """
    groups, keys_as_int = np.unique(arr, return_inverse = True)
    n_keys = max(keys_as_int)
    indices = [[] for i in range(n_keys + 1)]
    for i, k in enumerate(keys_as_int):
        indices[k].append(i)
    indices = [np.array(elt) for elt in indices]
    return dict(zip(groups, indices))




import argparse
from concurrent.futures import ProcessPoolExecutor, as_completed
import collections
import functools
from itertools import tee
import logging
import os
from timeit import default_timer as now
from uuid import uuid4

import h5py
import numpy as np

from fast5_research.fast5 import Fast5, iterate_fast5
from fast5_research.fast5_bulk import BulkFast5
from fast5_research.util import _sanitize_data_for_writing, readtsv, group_vector

def triplewise(iterable):
    a, b, c = tee(iterable, 3)
    next(b)
    next(c)
    next(c)
    yield from zip(a, b, c)

def extract_read_summary():
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO
    )
    logger = logging.getLogger('Summarize Reads')
    parser = argparse.ArgumentParser(description='Summarize reads stored in a Bulk .fast5')
    parser.add_argument('input', help='Bulk .fast5 file for input.')
    parser.add_argument('output', help='Output text file.')
    parser.add_argument('--channel_range', nargs=2, type=int, default=None, help='Channel range (inclusive).')
    args = parser.parse_args()

    if args.channel_range is None:
        with BulkFast5(args.input) as src:
            channels = src.channels
    else:
        channels = range(args.channel_range[0], args.channel_range[1] + 1)

    with BulkFast5(args.input) as src, open(args.output, 'w') as out_fh:
        extract_read_summary_internal(src, channels, out_fh, logger)


def extract_read_summary_internal(src, channels, out_fh, logger):
    fields = [
        'run', 'channel', 'mux', 'read_id', 'classification',
        'start_time', 'duration', 'time_since_strand',
        'start_event', 'end_event', 'num_events',
        'median_current', 'median_current_before', 'median_current_after',
        'drift', 'median_sd', 'range_current', 'median_dwell'
    ]
    out_fh.write('\t'.join(fields))
    out_fh.write('\n')
    sample_rate = src.sample_rate
    run_id = src.get_tracking_meta()['run_id']
    for chan in channels:
        logger.info("Processing channel {}".format(chan))
        count = 0
        reads = src.get_reads(chan)
        last_strand = 0
        # this drops the first and last read
        for before, data, after in triplewise(reads):
            data['run'] = run_id
            data['channel'] = chan
            data['mux'] = src.get_mux(chan, raw_index=data['read_start'] + data['read_length'] // 2)
            data['median_current'] = data['median']
            data['median_current_before'] = before['median']
            data['median_current_after'] = after['median']
            data['num_events'] = data['event_index_end'] - data['event_index_start']
            data['start_time'] = float(data['read_start']) / sample_rate
            data['duration'] = float(data['read_length']) / sample_rate
            data['start_event'] = data['event_index_start']
            data['end_event'] = data['event_index_end']
            data['range_current'] = data['range']
            data['median_dwell'] /= sample_rate
            data['time_since_strand'] = data['start_time'] - last_strand
            out_fh.write('\t'.join(str(data[f]) for f in fields))
            out_fh.write('\n')
            if data['classification'] == 'strand':
                last_strand = data['start_time'] + data['duration']
            count += 1
        logger.info("{} reads in channel {}".format(count, chan))



def extract_reads():
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO
    )
    logger = logging.getLogger('Extract Reads')
    parser = argparse.ArgumentParser(description='Bulk .fast5 to read .fast5 conversion.')
    parser.add_argument('input', help='Bulk .fast5 file for input.')
    parser.add_argument('output', help='Output folder.')
    out_format = parser.add_mutually_exclusive_group()
    out_format.add_argument('--multi', action='store_true', help='Output multi-read files.')
    out_format.add_argument('--single', action='store_false', dest='multi', help='Output single-read files.')
    parser.add_argument('--flat', default=False, action='store_true',
                        help='Create all .fast5 files in one directory')
    parser.add_argument('--by_id', default=False, action='store_true',
                        help='Name single-read .fast5 files by read_id.')
    parser.add_argument('--prefix', default="", help='Read file prefix.')
    parser.add_argument('--channel_range', nargs=2, type=int, default=None, help='Channel range (inclusive).')
    parser.add_argument('--summary', help='Strand summary file containing at least columns channel, start_time and duration).')
    parser.add_argument('--workers', type=int, default=4, help='Number of worker processes.')
    parser.add_argument('--limit', type=int, default=None, help='Limit reads per channel.')
    args = parser.parse_args()

    if not os.path.exists(args.output):
        os.makedirs(args.output)
    else:
        raise IOError('The output directory must not exist.')

    if args.summary is not None:
        if not os.path.isfile(args.summary):
            raise IOError('The summary file does not exist.')
        else:
            # load summary
            args.summary = np.genfromtxt(args.summary, delimiter='\t', encoding=None, dtype=None, names=True)

    worker = functools.partial(
        extract_channel_reads,
        args.input, args.output, args.prefix, args.flat, args.by_id,
        args.limit, args.multi
    )

    if args.channel_range is None:
        with BulkFast5(args.input) as src:
            channels = src.channels
    else:
        channels = range(args.channel_range[0], args.channel_range[1] + 1)

    if args.summary is not None:
        # only process channels in the summary
        summ_channels = set(args.summary['channel'])
        channels = [ch for ch in channels if ch in summ_channels]
        summary_by_ch = {ch: args.summary[np.where(args.summary['channel'] == ch)] for ch in channels}
    else:
        summary_by_ch = collections.defaultdict(lambda: None)

    if args.workers > 1:
        with ProcessPoolExecutor(args.workers) as executor:
            futures = [executor.submit(worker, c, summary=summary_by_ch[c]) for c in channels]
            for future in as_completed(futures):
                try:
                    n_reads, channel = future.result()
                except Exception as e:
                    logger.warning("Error processing channel.")
                    print(e)
                else:
                    logger.info("Extracted {} reads from channel {}.".format(n_reads, channel))
    else:
        for channel in channels:
            worker(channel, summary=summary_by_ch[channel])
    logger.info("Finished.")


def time_cast(time, sample_rate):
    """
    Convert a float time to sample index, or return time unmodified
    """
    if isinstance(time, float):
        return int(time * sample_rate)
    else:
        return time


def extract_channel_reads(source, output, prefix, flat, by_id, max_files, multi, channel, summary=None):
    if flat:
        out_path = output
        # give multi files a channel prefix else they will
        # conflict between channels. Singles already get
        # a "ch" component in their name
        if multi:
            extra = 'ch{}'.format(channel)
            if prefix == '':
                prefix = extra
            else:
                prefix = '{}_{}'.format(prefix, extra)
    else:
        out_path = os.path.join(output, str(channel))
        os.makedirs(out_path)

    with BulkFast5(source) as src:
        meta = src.get_metadata(channel)
        tracking_id = src.get_tracking_meta()
        context_tags = src.get_context_meta()
        channel_id = {
            'channel_number': channel,
            'range': meta['range'],
            'digitisation': meta['digitisation'],
            'offset': meta['offset'],
            'sampling_rate': meta['sample_rate']
        }

        Writer = MultiWriter if multi else SingleWriter
        with Writer(out_path, by_id, prefix=prefix) as writer:

            median_before = None
            counter = 1
            raw_data = src.get_raw(channel, use_scaling=False)

            if summary is not None:
                # convert array into stream of dicts
                reads = ({k: row[k] for k in row.dtype.names} for row in summary)
                class_field = 'class'
                start_field = 'start_time'
                duration_field = 'duration'
                # if start_time is a float (seconds) we need to convert to
                # samples
                time_cols = ['start_time', 'duration']
            else:
                reads = src.get_reads(channel)
                class_field = 'classification'
                start_field = 'read_start'
                duration_field = 'read_length'

            for read_number, read in enumerate(reads):

                if summary is not None:
                    if 'median_current_before' in read:
                        median_before = read['median_current_before']
                    else:
                        median_before = 0.0
                elif median_before is None:
                    median_before = read['median']
                    continue

                if summary is None and read[class_field] != 'strand':
                    median_before = read['median']
                else:
                    counter += 1
                    start = time_cast(read[start_field], meta['sample_rate'])
                    length = time_cast(read[duration_field], meta['sample_rate'])
                    read_id = {
                        'start_time': start,
                        'duration': length,
                        'read_number': read_number,
                        'start_mux': src.get_mux(channel, raw_index=start, wells_only=True),
                        'read_id': str(read['read_id']) if 'read_id' in read else str(uuid4()),
                        'scaling_used': 1,
                        'median_before': median_before
                    }

                    raw_slice = raw_data[start:start+length]
                    read = Read(read_id, read_number, tracking_id, channel_id, context_tags, raw_slice)
                    writer.write_read(read)
                    if counter == max_files:
                        break
    return counter, channel


def build_read_index():
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO
    )
    logger = logging.getLogger('Index Reads')

    parser = argparse.ArgumentParser(description='Build index of reads within .fast5s. Output to stdout.')
    parser.add_argument('input', help='.fast5 directory')
    parser.add_argument('--recursive', action='store_true',
        help='Search recursively under `input` for source files.')
    parser.add_argument('--workers', type=int, default=8,
        help='Number of worker processes.')
    args = parser.parse_args()

    src_files = list(iterate_fast5(args.input, paths=True, recursive=args.recursive))
    logger.info("Found {} files.".format(len(src_files)))

    with ProcessPoolExecutor(args.workers) as executor:
         n_reads = 0
         for i, (src, read_ids) in enumerate(
                 zip(src_files, executor.map(reads_in_multi, src_files, chunksize=10))):
             n_reads += len(read_ids)
             for read in read_ids:
                 print('\t'.join((read, os.path.abspath(src))))
             if i % 10 == 0:
                 logger.info("Indexed {}/{} files. {} reads".format(i, len(src_files), n_reads))


def filter_multi_reads():
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO
    )
    logger = logging.getLogger('Filter')
    parser = argparse.ArgumentParser(
        description='Extract reads from multi-read .fast5 files.')
    parser.add_argument('input',
        help='Path to input multi-read .fast5 files (or list of files).')
    parser.add_argument('output',
        help='Output folder.')
    parser.add_argument('filter',
        help='A .tsv file with column `read_id` defining required reads. '
             'If a `filename` column is present, this will be used as the '
             'location of the read.')
    parser.add_argument('--tsv_field', default='read_id',
        help='Field name from `filter` file to obtain read IDs.')
    parser.add_argument('--prefix', default="",
        help='Read file prefix.')
    parser.add_argument('--recursive', action='store_true',
        help='Search recursively under `input` for source files.')
    parser.add_argument('--workers', type=int, default=4,
        help='Number of worker processes.')

    out_format = parser.add_mutually_exclusive_group()
    out_format.add_argument('--multi', action='store_true', default=True,
        help='Output multi-read files.')
    out_format.add_argument('--single', action='store_false', dest='multi',
        help='Output single-read files.')

    #parser.add_argument('--limit', type=int, default=None, help='Limit reads per channel.')
    args = parser.parse_args()

    if not args.multi:
        raise NotImplementedError('Extraction of reads to single read files is on the TODO list.')

    if not os.path.exists(args.output):
        os.makedirs(args.output)
    else:
        raise IOError('The output directory must not exist.')

    # grab list of source files
    logger.info("Searching for input files.")
    try:
        src_files = list(set(readtsv(args.input)['filename']))
    except Exception as e:
        logger.info('Failed to read `input` as filelist, assuming path to search. {}'.format(e))
        src_files = list(iterate_fast5(args.input, paths=True, recursive=args.recursive))
    n_files = len(src_files)
    logger.info("Found {} source files.".format(n_files))

    logger.info("Reading filter file.")
    read_table = readtsv(args.filter, fields=[args.tsv_field])
    logger.info("Found {} reads in filter.".format(len(read_table)))

    try:
        # try to build index from the filter file with 'filename' column
        if 'filename' not in read_table.dtype.names:
            raise ValueError("'filename' column not present in filter.")
        logger.info("Attempting to build read index from input filter.")
        src_path_files = {
            os.path.basename(x):x for x in src_files
        }
        if len(src_path_files) != len(src_files):
            raise ValueError('Found non-uniquely named source files')
        read_index = dict()
        for fname, indices in group_vector(read_table['filename']).items():
            fpath = src_path_files[os.path.basename(fname)]
            read_index[fpath] = read_table[args.tsv_field][indices]
        logger.info("Successfully build read index from input filter.")
    except Exception as e:
        logger.info("Failed to build read index from summary: {}".format(e))
        read_index = None
        required_reads = set(read_table[args.tsv_field])
        logger.info("Finding reads within {} source files.".format(n_files))
        index_worker = functools.partial(reads_in_multi, filt=required_reads)
        read_index = dict()
        n_reads = 0
        with ProcessPoolExecutor(args.workers) as executor:
            i = 0
            for src_file, read_ids in zip(src_files, executor.map(index_worker, src_files, chunksize=10)):
                i += 1
                n_reads += len(read_ids)
                read_index[src_file] = read_ids
                if i % 10 == 0:
                    logger.info("Indexed {}/{} files. {}/{} reads".format(i, n_files, n_reads, len(required_reads)))

    n_reads = sum(len(x) for x in read_index.values())
    # We don't go via creating Read objects, copying the data verbatim
    # likely quicker and nothing should need the verification that the APIs
    # provide (garbage in, garbage out).
    logger.info("Extracting {} reads.".format(n_reads))
    if args.prefix != '':
        args.prefix = '{}_'.format(args.prefix)

    with ProcessPoolExecutor(args.workers) as executor:
        reads_per_process = np.ceil(n_reads / args.workers)
        proc_n_reads = 0
        proc_reads = dict()
        job = 0
        futures = list()
        for src in read_index.keys():
            proc_reads[src] = read_index[src]
            proc_n_reads += len(proc_reads[src])
            if proc_n_reads > reads_per_process:
                proc_prefix = "{}{}_".format(args.prefix, job)
                futures.append(executor.submit(_subset_reads_to_file, proc_reads, args.output, proc_prefix, worker_id=job))
                job += 1
                proc_n_reads = 0
                proc_reads = dict()
        if proc_n_reads > 0:  # processing remaining reads
            proc_prefix = "{}{}_".format(args.prefix, job)
            futures.append(executor.submit(_subset_reads_to_file, proc_reads, args.output, proc_prefix, worker_id=job))


        for fut in as_completed(futures):
            try:
                reads_written, prefix = fut.result()
                logger.info("Written {} reads to {}.".format(reads_written, prefix))
            except Exception as e:
                logger.warning("Error: {}".format(e))
    logger.info("Done.")


def _subset_reads_to_file(read_index, output, prefix, worker_id=0):
    logger = logging.getLogger('Worker-{}'.format(worker_id))
    n_reads = sum(len(x) for x in read_index.values())
    reads_written = 0
    t0 = now()
    with MultiWriter(output, None, prefix=prefix) as writer:
        for src_file, read_ids in read_index.items():
            reads_written += len(read_ids)
            t1 = now()
            if t1 - t0 > 30: # log update every 30 seconds
                logger.info("Written {}/{} reads ({:.0f}% done)".format(
                    reads_written, n_reads, 100 * reads_written / n_reads
                ))
                t0 = t1
            with h5py.File(src_file, 'r') as src_fh:
                for read_id in read_ids:
                    try:
                        read_grp = src_fh["read_{}".format(read_id)]
                    except:
                        logger.warning("Did not find {} in {}.".format(read_id, src_fh.filename))
                    else:
                        writer.write_read(read_grp)
    return reads_written, prefix


def reads_in_multi(src, filt=None):
    """Get list of read IDs contained within a multi-read file.

    :param src: source file.
    :param filt: perform filtering by given set.
    :returns: set of read UUIDs (as string and recorded in hdf group name).
    """
    logger = logging.getLogger(os.path.splitext(os.path.basename(src))[0])
    logger.debug("Finding reads.")
    prefix = 'read_'
    with h5py.File(src, 'r') as fh:
        read_ids = set(grp[len(prefix):] for grp in fh if grp.startswith(prefix))
    logger.debug("Found {} reads.".format(len(read_ids)))
    if filt is not None:
        read_ids = read_ids.intersection(filt)
    logger.debug("Filtered to {} reads.".format(len(read_ids)))
    return read_ids


class Read(object):
    # Just a sketch to help interchange of format
    def __init__(self, read_id, read_number, tracking_id, channel_id, context_tags, raw):
        self.read_id = read_id
        self.read_number = read_number
        self.tracking_id = tracking_id
        self.channel_id = channel_id
        self.context_tags = context_tags
        self.raw = raw

        # ensure typing and required fields
        self.channel_id = Fast5.convert_channel_id(self.channel_id)
        self.tracking_id = Fast5.convert_tracking_id(self.tracking_id)


class ReadWriter(object):
    def __init__(self, out_path, by_id, prefix=""):
        self.out_path = out_path
        self.by_id = by_id
        if prefix != "":
            prefix = "{}_".format(prefix)
        self.prefix = prefix

    def write_read(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass


class SingleWriter(ReadWriter):
    def write_read(self, read):
        if self.by_id:
            filename = '{}.fast5'.format(read.read_id['read_id'])
        else:
            filename = '{}read_ch{}_file{}.fast5'.format(
                self.prefix, read.channel_id['channel_number'], read.read_number
            )
        filename = os.path.join(self.out_path, filename)
        with Fast5.New(filename, 'a', tracking_id=read.tracking_id, context_tags=read.context_tags, channel_id=read.channel_id) as h:
            h.set_raw(read.raw, meta=read.read_id, read_number=read.read_number)


MULTI_READ_FILE_VERSION = "2.0"

class MultiWriter(ReadWriter):
    def __init__(self, out_path, by_id, prefix="", reads_per_file=4000):
        super(MultiWriter, self).__init__(out_path, by_id, prefix=prefix)
        self.reads_per_file = reads_per_file
        self.current_reads = 0 # reads in open file, used to signal new file condition
        self.file_counter = 0
        self.current_file = None
        self.closed = False


    def __exit__(self, exception_type, exception_value, traceback):
        self.close()


    def close(self):
        if isinstance(self.current_file, h5py.File):
            self.current_file.close()


    def write_read(self, read):
        """Write a read.

        :param read: either a `Read` object or an hdf group handle from a
            source multi-read file.
        """
        if self.closed:
            raise RuntimeError('Cannot write after closed.')

        if self.current_reads == 0:
            # start a new file
            self.close()
            filename = '{}mreads_file{}.fast5'.format(
                self.prefix, self.file_counter
            )
            filename = os.path.join(self.out_path, filename)
            self.current_file = h5py.File(filename, 'w')
            self.current_file.attrs[_sanitize_data_for_writing('file_version')] = _sanitize_data_for_writing("2.0")
            self.file_counter += 1

        # write data
        if isinstance(read, Read):
            self._write_read(read)
        elif isinstance(read, h5py.Group):
            self._copy_read_group(read)
        else:
            raise TypeError("Cannot write type {} to output file.")
        self.current_reads += 1

        # update
        if self.current_reads == self.reads_per_file:
            self.current_reads = 0


    def _write_read(self, read):
        if read.raw.dtype != np.int16:
            raise TypeError('Raw data must be of type int16.')

        read_group = '/read_{}'.format(read.read_id['read_id'])
        Fast5._add_attrs_to_fh(self.current_file, {'run_id': read.tracking_id['run_id']}, read_group, convert=str)

        # add all attributes
        for grp_name in ('tracking_id', 'context_tags'):
            # spec has all of these as str
            data = getattr(read, grp_name)
            Fast5._add_attrs_to_fh(self.current_file, data, '{}/{}'.format(read_group, grp_name), convert=str)
        Fast5._add_attrs_to_fh(self.current_file, read.channel_id, '{}/channel_id'.format(read_group))

        # add the data (and some more attrs)
        data_path = '{}/Raw'.format(read_group)
        read_id = Fast5._convert_meta_times(read.read_id, read.channel_id['sampling_rate'])
        read_id = Fast5.convert_raw_meta(read_id)
        Fast5._add_attrs_to_fh(self.current_file, read_id, data_path)
        signal_path = '{}/Signal'.format(data_path)
        self.current_file.create_dataset(
            signal_path, data=read.raw, compression='gzip', shuffle=True, dtype='i2')


    def _copy_read_group(self, read):
        self.current_file.copy(read, read.name)



import os
import re
import sys
from setuptools import setup, find_packages


__pkg_name__ = 'fast5_research'
__author__ = 'cwright'
__description__ = 'ONT Research .fast5 read/write API.'

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
with open(path.join(this_directory, 'README.md'), **kwargs) as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'


# Get the version number from __init__.py
verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))

dir_path = os.path.dirname(__file__)
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    install_requires = [
        r.split('#')[0].strip()
        for r in fh.read().splitlines() if not r.strip().startswith('#')
    ]

extra_requires={}

py2only_requirements = ['futures']
if len(py2only_requirements) > 0:
    extra_requires[':python_version == "2.7"'] = []

for requirement in py2only_requirements:
    install_requires.remove(requirement)
    extra_requires[':python_version == "2.7"'].append(requirement)


setup(
    name=__pkg_name__,
    version=__version__,
    url='https://github.com/nanoporetech/{}'.format(__pkg_name__),
    author=__author__,
    author_email='{}@nanoporetech.com'.format(__author__),
    description=__description__,
    long_description=__long_description__,
    long_description_content_type=__long_description_content_type__,
    entry_points={
        'console_scripts': [
            'index_reads = {}.extract:build_read_index'.format(__pkg_name__),
            'extract_reads = {}.extract:extract_reads'.format(__pkg_name__),
            'read_summary = {}.extract:extract_read_summary'.format(__pkg_name__),
            'filter_reads = {}.extract:filter_multi_reads'.format(__pkg_name__),
        ]
    },
    license='Mozilla Public License 2.0',
    dependency_links=[],
    install_requires=install_requires,
    tests_require=['nose>=1.3.7'].extend(install_requires),
    extras_require=extra_requires,
    packages=find_packages(exclude=['*.test', '*.test.*', 'test.*', 'test']),
    package_data={},
    zip_safe=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    keywords='ONT Research fast5 API',
)
